import logging
import os
from pathlib import Path, PosixPath

import lightning as L
import torch
import torch.utils.data as data
from lightning.pytorch.loggers import WandbLogger
from lightning.pytorch.tuner.tuning import Tuner
from transformers import AutoTokenizer

from .src.dataset import DataModule
from .src.lightningMods import LightningMod_DocEmbedder
from .src.utils.general import getargs, upload_checkpoint_to_gcs

torch.set_float32_matmul_precision("medium")

args = getargs()

# Create Logger For file stream and console stream
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

sh = logging.StreamHandler()
fh = logging.FileHandler("main.log")

# Create Formatting
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
sh.setFormatter(formatter)
fh.setFormatter(formatter)

logger.addHandler(sh)
logger.addHandler(sh)

args_dict = vars(args)

# Seed everything
L.seed_everything(args.seed)

# Setup google Credentials gor GCS
os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = args.gcs_credentials
logger.info(
    "🗝 Using GCS Credentials (for uploading model checkpoint): {}".format(
        args.gcs_credentials
    )
)

# Create wandb logger
wandb_logger = None
if args.wandb:
    logger.info("🪄Setting up wandb")
    wandb_logger = WandbLogger(project="SciBertPatentRanker")


# 🤖 Create Model
device = torch.device("cuda")
tokenizer = AutoTokenizer.from_pretrained(
    args.encoder_base,
)

# Form psql_args:
psql_args = {
    "host": args.psql_host,
    "database": args.psql_database,
    "port": args.psql_port,
    "user": args.psql_user,
    "password": args.psql_password,
    "relationships_table": args.psql_relationships_tablename,
    "samples_text": args.psql_reltexts_tablename,
}


###################################
# Create Training Environment
###################################
trainer = L.Trainer(
    accelerator="gpu",
    devices=1,
    logger=wandb_logger,
    accumulate_grad_batches=4,
    log_every_n_steps=1,
    max_epochs=args.epochs,
    # val_check_interval=0.1,
    enable_checkpointing=True,
)
logger.info("Passing the training to ⚡Pytorch-Lightning")
# Load Pytorch Lightning Modules
# Check if checkpoint exists
if Path("checkpoints/model.ckpt").exists():
    logger.info("Loading checkpoint")
    model = LightningMod_DocEmbedder.load_from_checkpoint(
        "checkpoints/model.ckpt",
        lang_head_name=args.encoder_base,
        tokenizer=tokenizer,
        load_init_weights=True,
    )
else:
    model = LightningMod_DocEmbedder(
        args.encoder_base, load_init_weights=True, tokenizer=tokenizer
    )
logger.info("🤖 Model Created")
###################################
# Prepare Data
###################################
# Create DataModule
dataModule = DataModule(
    args.batch_size,
    psql_args,
    args.data_path,
    tokenizer,
    args.encoder_max_length,
    device=device,
)
# dataModule.prepare_data()
tuner = Tuner(trainer)
tuner.scale_batch_size(model, mode="binsearch", datamodule=dataModule)
###################################
# Train
###################################
trainer.fit(model, datamodule=dataModule)
# Trainer save weigths
trainer.save_checkpoint("checkpoints/model.ckpt")
logger.info("⏫ Uploading Trained model to CGS")
upload_checkpoint_to_gcs(model, args.gcs_bucketname, args.gcs_blobname)



import logging
import os

import torch
from google.cloud import storage
from tqdm import tqdm
from transformers import AutoModel, AutoTokenizer, PreTrainedTokenizer

from .src.lightningMods import LightningMod_DocEmbedder
from .src.utils.general import setup_logger


class DocRep:
    def __init__(self, title, abstract):
        self.title = title
        self.abstract = abstract
        self.data = "<title>" + self.title + "[SEP]<abstract>" + self.abstract


class DocEmbedder:
    BUCKET_NAME: str = "llmbi-models"
    TARGET_PATH: str = ".llm4bi/"
    FOUNDATION_MODEL = "allenai/scibert_scivocab_uncased"
    ENCODER_MAX_LENGTH = 512

    def __init__(self, blob_name: str):
        self.logger = setup_logger("DocEmbWrapper", "main.log", logging.INFO)
        home_path = os.path.expanduser("~")
        target_dir = os.path.join(home_path, self.TARGET_PATH)
        target_file: str = os.path.join(target_dir, blob_name)
        full_dir = os.path.dirname(target_file)

        self.tokenizer = AutoTokenizer.from_pretrained(self.FOUNDATION_MODEL)

        os.makedirs(full_dir, exist_ok=True)
        self.blob_name = blob_name

        # Check for cache file
        if os.path.exists(target_file):
            self.logger.info(
                f"Cache found. Loading your model {blob_name} from dir {target_dir}"
            )
        else:
            # Load from CGS
            self.logger.info(
                f"Cache not found. Downloading your model {blob_name} from GCS into dir {target_dir}"
            )
            self._dload_from_gcs(target_file)
            self.logger.info("Download complete")
        # Load into lightining
        self.model = LightningMod_DocEmbedder.load_from_checkpoint(
            target_file,
            lang_head_name=self.FOUNDATION_MODEL,
            tokenizer=self.tokenizer,
            load_init_weights=False,
            strict=False,
        ).to("cuda")

    def _dload_from_gcs(self, target_file: str) -> None:
        if os.environ.get("GOOGLE_APPLICATION_CREDENTIALS") is None:
            # Ask for path from user
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = input(
                "Enter path to service key: "
            )
        storage_client = storage.Client()
        bucket = storage_client.bucket(self.BUCKET_NAME)
        blob = bucket.blob(self.blob_name)
        with open(target_file) as f:
            with tqdm.wrapattr(f, "write", total=blob.size) as file_obj:
                storage_client.download_blob_to_file(blob, file_obj)
        # blob.download_to_filename(target_file)

    def __call__(self, doc: DocRep) -> torch.Tensor:
        return self.model(doc.data)

    def doc_dist(self, doc1: DocRep, doc2: DocRep) -> torch.Tensor:
        self.model.eval()
        tokns1 = (
            torch.Tensor(
                self.tokenizer.encode(
                    doc1.data,
                    max_length=self.ENCODER_MAX_LENGTH,
                    truncation=True,
                    padding="max_length",
                )
            )
            .to(torch.long)
            .view(1, -1)
            .to("cuda")
        )
        attention_mask = torch.ones_like(tokns1)
        attention_mask[tokns1 == 0] = 0
        embed1 = (
            self.model(tokns1, attention_mask=attention_mask)
            .hidden_states[-1][:, 0, :]
            .detach()
            .squeeze()
        )
        tokns2 = (
            torch.Tensor(
                self.tokenizer.encode(
                    doc2.data,
                    max_length=self.ENCODER_MAX_LENGTH,
                    truncation=True,
                    padding="max_length",
                )
            )
            .to(torch.long)
            .view(1, -1)
            .to("cuda")
        )
        attention_mask = torch.ones_like(tokns2).to("cuda")
        attention_mask[tokns2 == 0] = 0
        embed2 = (
            self.model(tokns2, attention_mask=attention_mask)
            .hidden_states[-1][:, 0, :]
            .detach()
            .squeeze()
        )

        return torch.norm(
            embed1 - embed2,
            p=2,
            dim=0,
            keepdim=True,
        )

    def dist(self, doc: DocRep, docRep):
        return



import os
from argparse import ArgumentParser
from pathlib import Path

from transformers import AutoTokenizer

from docembedder import DocEmbedder, DocRep
from src.lightningMods import LightningMod_DocEmbedder

# Key location
os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = "gcs_service_key.json"
ap = ArgumentParser()
ap.add_argument("--path", type=str, default="checkpoints/model.ckpt")
args = ap.parse_args()

tokenizer = AutoTokenizer.from_pretrained("allenai/scibert_scivocab_uncased")

# Original file, related file, and non-related file
og_file_title = "USB connection-detection circuitry and operation methods of the same"
og_file_abstract = "A USB connection-detection circuitry and the operation method of the same are disclosed. The circuitry includes a transmitting circuit and a detecting circuit. The transmitting circuit contains a pair of differential signal lines, a pair of pull-down resistors and a pair of pull-up resistors wherein one pull-down resistor and one pull-up resistor are connected to the same differential signal line with their own individidual"

related_file_title = "Microcontroller input/output nodes with both programmable pull-up and pull-down resistive loads and programmable drive strength "
related_file_abstract = "The present invention relates to an input/output node in an electronic device which comprises an input/output pin, a plurality of programmable pull-up resistors and a plurality of programmable pull-down resistors. Each of the pull-up and pull-down resistors, or a combination of them, can be activated by turning on or off n-MOS and p-MOS transistors with logic c"

non_related_file_title = "Reusable pesticide bait station"
non_related_file_abstract = "A compact reusable bait station that can be used for either solid or liquid bait. The bait station has an outer wall that defines an inner bait chamber. The chamber, when sealed, prevents liquid bait inside the chamber from flowing out. The chamber has limited access from the outside to reduce evaporation, drying, and contamination of the liquid bait by preventing unnecessary exposure to the environment. Access to the outside is limited to access ports th"

# Create DocRep objects
og_doc = DocRep(og_file_title, og_file_abstract)
related_doc = DocRep(related_file_title, related_file_abstract)
non_related_doc = DocRep(non_related_file_title, non_related_file_abstract)

# Create DocEmbedder object
if Path(args.path).exists():
    doc_embedded = DocEmbedder(args.path, tokenizer)

    print(
        "Distance between original and related:",
        doc_embedded.doc_dist(og_doc, related_doc),
    )
    print(
        "Distance between original and non-related:",
        doc_embedded.doc_dist(og_doc, non_related_doc),
    )
else:
    raise ValueError("Invalid path")

# Get distances between documents

# Thats pretty much it



import argparse
import logging
import os

import lightning as L
from google.cloud import storage


def getargs():
    argparser = argparse.ArgumentParser(
        description="PyTorch implementation of Document Embeddings"
    )

    # General Arguments
    argparser.add_argument(
        "--data-path",
        type=str,
        default="dataset/",
        help="Location of the data corpus, already stored as a parquet file",
    )
    argparser.add_argument("--epochs", type=int, default=1, help="Number of epochs")
    argparser.add_argument("--batch_size", type=int, default=4, help="Batch size")
    argparser.add_argument("--lr", type=float, default=0.001, help="Learning rate")
    argparser.add_argument("--seed", type=int, default=420, help="Random seed")
    argparser.add_argument("--cuda", action="store_true", help="use CUDA")
    argparser.add_argument("--wandb", action="store_true", help="Report to wandb")
    argparser.add_argument(
        "--save", type=str, default="model.pt", help="Path to save the final model"
    )
    argparser.add_argument(
        "--load", type=str, default="", help="Path to load the model"
    )
    argparser.add_argument(
        "--log_interval", type=int, default=200, help="report interval"
    )
    argparser.add_argument(
        "--encoder_base",
        type=str,
        default="allenai/scibert_scivocab_uncased",
        help="Encoder base",
    )
    argparser.add_argument(
        "--encoder_max_length",
        type=int,
        default=512,
        help="Encoder max length",
    )
    # For CGS
    argparser.add_argument(
        "--gcs_credentials",
        default="gcs_service_key.json",
        help="Credentials for Google Cloud Storage",
    )
    argparser.add_argument("--gcs_bucketname", default="llmbi-models")
    argparser.add_argument("--gcs_blobname", default="checkpoints/model.ckpt")

    # Arguments for psql
    argparser.add_argument("--psql_host", required=True)
    argparser.add_argument("--psql_database", required=True)
    argparser.add_argument("--psql_user", required=True)
    argparser.add_argument("--psql_password", required=True)
    argparser.add_argument("--psql_port", required=True)

    # Table Specific
    argparser.add_argument("--psql_relationships_tablename", default="samples_large")
    argparser.add_argument("--psql_reltexts_tablename", default="samples_large_text")

    return argparser.parse_args()


def setup_logger(name, file="main.log", level=logging.INFO):
    os.makedirs("~/.llm4bi/logs", exist_ok=True)
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    fh = logging.FileHandler(os.path.join("~/.llm4bi/logs", file), "w")
    sh = logging.StreamHandler()
    fh.setFormatter(formatter)
    sh.setFormatter(formatter)
    fh.setLevel(level)
    sh.setLevel(logging.INFO)

    logger.addHandler(fh)
    logger.addHandler(sh)
    return logger


def upload_checkpoint_to_gcs(model: L.LightningModule, bucketname, blob_name):
    local_path = "checkpoints/model.ckpt"

    # Initialize GCS clien
    storage_client = storage.Client()
    bucket = storage_client.bucket(bucketname)
    blob = bucket.blob(blob_name)

    # Upload checkpoint to CGS
    blob.upload_from_filename(local_path)



import logging
import random
from pathlib import Path, PosixPath
from typing import Dict, Generator, List, Tuple
from urllib.parse import quote_plus

import lightning as L
import numpy as np
import pandas as pd
import torch
import torch.utils.data as data
from torch import Tensor
from torch.utils.data import Dataset
from tqdm import tqdm
from transformers import PreTrainedTokenizer

from .utils.general import setup_logger

"""
Every Instance will contain:
    1. A source document
    2. One of its (possibly many) 1st-degree references
    3. One of its (possibly many) 2nd-degree references
    4. One of its (possibly many) 3rd-degree references
    5. One of its (possibly many) non-related documents
Each of them will receive an embedding and the external loss will be calculated
so as to create an appropriate margin between the source document and its references
"""


class DataModule(L.LightningDataModule):
    def __init__(
        self,
        batch_size,
        psql_args,
        data_path,
        tokenizer,
        encoder_max_length,
        device,
    ):
        super().__init__()
        self.logger = setup_logger(__name__, level=logging.INFO)
        self.save_hyperparameters()
        self.batch_size = batch_size

        self.device = device
        self.psql_args = psql_args
        self.data_path = data_path
        self.tokenizer = tokenizer
        self.encoder_max_length = encoder_max_length

    def prepare_data(self):
        self.logger.info("📂Creating Dataset")
        datasetFactory = DatabaseFactory(
            dir_to_cache=PosixPath(self.data_path),
            psql_args=self.psql_args,
            tokenizer=self.tokenizer,
            encoder_max_length=self.encoder_max_length,
        )
        self.train_dataset, self.val_dataset = datasetFactory.createDataset(self.device)

    def train_dataloader(self):
        return data.DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,  # type: ignore
            num_workers=12,
        )

    def val_dataloader(self):
        return data.DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,  # type: ignore
            num_workers=12,
        )


class DocumentDataset(Dataset):
    def __init__(
        self,
        samples: List[List[Tensor]],
    ):
        """
        Args:
            data_path: Path to the dataset
            samples: List of rows. Each rows contains n lists. Each list being the tokens corresponding to title and abstract of each patent.
            device: device to put the tensors on
        """
        self.logger = setup_logger("DocumentDataset", level=logging.DEBUG)
        self.dataset = samples

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        # return sel.dataset[idx] where dataset is dataframe
        return self.dataset[idx]


"""
This class will be used to pull data from the database and construct the dataset
args:
---
- psql_args: Dict[str, str] = {
        "host": Host where psql dataset is hosted,
        "database": Database name,
        "user": User name,
        "password": Password,
    }
- encoder_max_length: The maximum length of the encoder we will use for training/production
"""


class DatabaseFactory:
    # Check if cache exists
    def __init__(
        self,
        dir_to_cache: PosixPath,
        psql_args,
        tokenizer: PreTrainedTokenizer,
        encoder_max_length: int = 512,
    ):
        # Setup logger
        self.logger = setup_logger(__name__, level=logging.INFO)
        self.psql_args = psql_args
        self.encoder_max_length = encoder_max_length
        self.tokenizer = tokenizer
        # Get absolute path without name of file
        self.path_to_cache = Path(dir_to_cache).parent
        self.train_path = dir_to_cache / "dataset_train.parquet"
        self.val_path = dir_to_cache / "dataset_val.parquet"

    def createDataset(
        self, device: torch.device
    ) -> Tuple[DocumentDataset, DocumentDataset]:
        if self.train_path.exists() and self.val_path.exists():
            # Load cache
            # TODO: MAybe make this less clunky. Feels like im doing redundant calls
            self.logger.info("Found cache file. Will construct from them")
            self.train, self.val = self._construct_from_cache()
            # Take List[List[array-like]] and transform it into List[List[Tensor]]
            self.train = [
                [torch.Tensor(np.array(doc)).to(torch.long) for doc in sample]
                for sample in self.train
            ]
            self.val = [
                [torch.Tensor(np.array(doc)).to(torch.long) for doc in sample]
                for sample in self.val
            ]

        else:
            # TODO: check if I have to pass return here to Tensors just like I do when I
            # construct from cache
            self.logger.info("Unable to find cache file. Will construct from scratch")
            self.train, self.val = self._construct_from_strach(self.psql_args)

        return (DocumentDataset(self.train), DocumentDataset(self.val))

    def _construct_from_cache(self) -> Tuple[List, List]:
        # Read arrow files:
        train = pd.read_parquet(self.train_path).values.tolist()
        val = pd.read_parquet(self.val_path).values.tolist()
        return train, val

    def _construct_from_strach(self, psql_args: Dict[str, str]):
        # Create dbToDT
        dbToDt = DatabaseToDataset(psql_args)

        # We might want to cache this bad boi
        samples = dbToDt.pull_data(
            relationships_tablename=psql_args["relationships_table"],
            samples_text_tablename=psql_args["samples_text"],
        )
        # samples_df.to_csv("samples.csv", sep="|") # 🐛For debugging
        self.logger.info(
            f"Obtained the {len(samples)} text samples. Will no tokenized them"
        )

        tokenized_samples = self._tokenize_samples(samples)
        self.logger.info(
            "Difference of ds after tokenization : %s",
            len(samples) - len(tokenized_samples),
        )

        # Do random split
        self.logger.info("Creating train and validation split")
        random.shuffle(tokenized_samples)
        train_size = int(len(tokenized_samples) * 0.8)
        train_ds = tokenized_samples[:train_size]
        val_ds = tokenized_samples[train_size:]
        # Save to parquet
        self.logger.info(
            "Saving to cache fiels {} and {}".format(self.train_path, self.val_path)
        )
        # TODO: maybe be more flexible with how we write the columns rather than hardcoding them
        columns = (
            ["og_pubnum"]
            + [f"ref{i}" for i in range(len(train_ds[0]) - 2)]
            + ["nonref"]
        )
        train_df = pd.DataFrame(train_ds, columns=columns)
        val_df = pd.DataFrame(val_ds, columns=columns)
        train_df.to_parquet(self.train_path)
        val_df.to_parquet(self.val_path)

        return train_ds, val_ds

    def _tokenize_samples(self, samples: List[List[str]]) -> List[List[Tensor]]:
        # Before we save to parquet we want to tokenize the samples
        tokenized_samples = []
        for sample in tqdm(samples, desc="Tokenizing samples"):
            # TODO: verify that this indeed truncates to max size of encoder
            encoded_sample = [
                self.tokenizer.encode(
                    samp,
                    max_length=self.encoder_max_length,
                    truncation=True,
                    padding="max_length",
                )
                for samp in sample
            ]
            tokenized_samples.append(encoded_sample)

        self.logger.info("Tokenized the samples. Will now save them to parquet")

        # Save samples: [[]*n] into a parquet file
        return tokenized_samples


class DatabaseToDataset:
    def __init__(self, psql_args: Dict):
        # Establish psql connection to database
        from sqlalchemy import create_engine

        self.logger = setup_logger(
            __name__, "DatabaseToDataset.log", level=logging.INFO
        )
        self.psql_args = psql_args

        self.engine = create_engine(
            f"postgresql://{psql_args['user']}:{quote_plus(psql_args['password'])}@{psql_args['host']}:{psql_args['port']}/{psql_args['database']}"
        )
        self.conn = self.engine.connect()
        # Catch OperationalError:

        self.avg_samples_extracted_from_row = 0  # 🪲
        self.tot_samples_added = 0  # 🪲

    def _obtain_txt_data_for_row(self, row, table_name):
        # Query the text table for the text of the sample_id
        to_return = []
        to_remove = []
        for sample_id in row:
            query = f"SELECT title,abstract FROM {table_name} WHERE publication_number = '{sample_id}'"
            res = pd.read_sql(query, self.conn)
            # res.dropna(axis=0, how="any", inplace=True)
            # Check
            if res.empty:
                # We have no text for this sample_id
                self.logger.debug(
                    "No matching row in text data for publication number %s",
                    sample_id,
                )
                to_remove.append(sample_id)
                continue
            title = res.iloc[0]["title"]
            abstract = res.iloc[0]["abstract"]
            self.logger.debug(f"Title {title} and abstract {abstract}")
            if (
                title == None
                or title == ""
                or abstract == None
                or abstract == ""
                or title == "NaN"
                or abstract == "NaN"
            ):
                to_remove.append(sample_id)
                continue
            # Ensure title + abstract is less than encoder_max_length
            to_return += ["<title>: " + title + "[SEP]<abstract>: " + abstract]
        return to_return, to_remove

    def _obtain_txt_data_for_element(self, id, table_name):
        query = (
            f"SELECT title,abstract FROM {table_name} WHERE publication_number = '{id}'"
        )
        res = pd.read_sql(query, self.conn)
        if res.empty:
            return None
        title = res.iloc[0]["title"]
        abstract = res.iloc[0]["abstract"]
        if title == None or title == "" or abstract == None or abstract == "":
            return None
        return "<title>: " + title + "[SEP]<abstract>:" + abstract

    def _pull_one_sample_from_row(
        self, row: pd.Series, samples_text_tablename: str
    ) -> Generator[List[str], None, int]:
        """
        Will take a pandas series as row and get as many samples as possible from it
        """
        # Get column 0 before the others
        col_0_text = self._obtain_txt_data_for_element(
            row["og_pubnum"], samples_text_tablename
        )
        if col_0_text == None:
            self.logger.warning("An og_pubnum was found without text")
            return 0

        ids_in_row_without_text = 0
        pubs_added = 0

        while True:
            text_entry: List[str] = [col_0_text]
            # HACK: remember that for now you are doing simple ref,pos,neg. If you want to dela with chain you have to change this code
            # for col in row[1:]: # For all columns
            for col in row.iloc[[1, -1]]:  # Only includes pos and neg references
                obtained_text = False
                while obtained_text == False:
                    col_list: List[str] = col  # type: ignore
                    # Check Stopping Criterion
                    if len(col_list) == 0:
                        # report statistics
                        column_reports = str([len(c) for c in row[1:]])
                        # HACK: remove once you dont need use for debugging
                        self.logger.debug(
                            f"Statistics Report for og_pubnum {row['og_pubnum']}:\n"
                            f"\tElements taken on each column {pubs_added}\n"
                            f"\tStill remaining {column_reports}.\n"
                            f"\tPubs without text {ids_in_row_without_text}.\n"
                        )
                        return pubs_added
                    id = col_list.pop()

                    text_res = self._obtain_txt_data_for_element(
                        id, samples_text_tablename
                    )
                    if text_res != None:
                        text_entry.append(text_res)
                        obtained_text = True
                    else:
                        ids_in_row_without_text += 1  # 🪲

            pubs_added += 1  # 🪲
            yield text_entry

    def pull_data(self, relationships_tablename, samples_text_tablename: str):
        """
        Main method that will pull data from the database and construct the dataset
        Arguments:
            relationships_tablename: The name of the table in the database that contains the network structure
            samples_text_tablename: The name of the table in the database that contains the text data indexed by publication number
        """
        query = f"""SELECT * FROM {relationships_tablename}
            WHERE array_length(ref0,1) < 12
            AND array_length(ref1,1) < 400
            AND array_length(ref2,1) < 1000
            AND array_length(neg_exs,1) > 0
            ORDER BY og_pubnum ASC;
        """  # TODO: not a big fan of this way of filtering, well have to fix it later
        try:
            relationships: pd.DataFrame = pd.read_sql_query(query, self.conn)
        except Exception as e:
            self.logger.error("❌ Unable to query database:" + str(e))
            exit(-1)

        # Remove all None values from the dataframe
        relationships.dropna(axis=0, how="any", inplace=True)
        # relationships.to_csv("relationships.csv", sep="|")  # 🪲
        samples = []
        # Get columns

        # From there we will select just a few to build the dataset (atleast for now)
        row_i = 0
        tqdm_obj = tqdm(total=len(relationships), desc="Iterating through rows.")
        ####################
        # Get Text
        ####################
        for _, row_series in relationships.iterrows():
            try:
                tqdm_obj.update(1)
                for value in self._pull_one_sample_from_row(
                    row_series, samples_text_tablename
                ):  # When _pull_one_sample_from_row returns none it will stop
                    samples.append(value)
            except StopIteration as e:
                row_i += 1
                result = e.value
                self.avg_samples_extracted_from_row = (
                    (row_i - 1) / row_i
                ) * self.avg_samples_extracted_from_row + (1 / row_i) * result
                self.logger.debug(
                    "Average samples extracted from row: %s",
                    self.avg_samples_extracted_from_row,
                )

        # 🪲 Save this for debugging
        values_df = pd.DataFrame(samples)
        values_df.to_csv("values.csv", sep="|")
        # At this point we have all the samples indices we need
        self.logger.info(f"Formed the index dataset with total size {len(samples)}.")
        return samples



import logging
from typing import List

import lightning as L
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.utils.data as data
from torch import Tensor
from transformers import AutoConfig, AutoModel, PreTrainedTokenizer

from .loss import MarginilizedLoss
from .utils.general import setup_logger


class LightningMod_DocEmbedder(L.LightningModule):
    def __init__(
        self,
        lang_head_name,
        tokenizer: PreTrainedTokenizer,
        load_init_weights=True,
    ):
        super().__init__()
        self.mylogger = setup_logger(
            "LightningMod_DocEmbedder",
            "LightningMod_DocEmbedder.log",
            level=logging.INFO,
        )
        self.lang_head_config = AutoConfig.from_pretrained(
            lang_head_name, output_hidden_states=True
        )
        # self.my_head = nn.Linear(
        # self.lang_head_config.hidden_size, self.lang_head_config.hidden_size
        # )
        # For batch_size finding
        if load_init_weights:  # Actually downloads the weights (from HF likely)
            self.mylogger.info("⏬Downloading Model")
            # Create config so we can get hidden ouputs
            config = AutoConfig.from_pretrained(
                lang_head_name, output_hidden_states=True
            )
            self.lang_head = AutoModel.from_pretrained(
                lang_head_name
            )  # , config=config)
        else:  # This model has the correct architecture but randomly initalized weights
            self.mylogger.info("🛠️ Configuring Model Architecture")
            self.lang_head = AutoModel.from_config(self.lang_head_config)

        # TODO: remove this from loss. Its was only for bug huntin 🐛
        self.loss = MarginilizedLoss(tokenizer)

    def forward(self, x, attention_mask):
        return self.lang_head(x, attention_mask=attention_mask)

    def training_step(self, ref_batches, batch_idx):
        # Get embeddings
        # self.mylogger.debug(f"ref_batches: {ref_batches}")
        total_norm = 0.0
        for p in self.lang_head.parameters():
            if p.grad is not None:
                param_norm = p.grad.data.norm(2)
                total_norm += param_norm.item() ** 2
        total_norm = total_norm ** (1.0 / 2)

        self.mylogger.debug(f"Total nurm at {batch_idx} is {total_norm}")
        # self.mylogger.debug(f"reg_batches: looks like {ref_batches})")
        contextualized_embeddings_batches = []
        self.mylogger.debug(f"Number of batches: {len(ref_batches)}")
        for batch in ref_batches:
            # Create attention mask online for the batch
            attention_mask = torch.ones_like(batch)
            attention_mask[batch == 0] = 0
            # cont_embeddings = self.lang_head(
            # batch, attention_mask=attention_mask
            # ).hidden_states[-1]
            # TODO: trying type_ids
            cont_embeddings = self.lang_head(batch, attention_mask=attention_mask)
            # lm_output = cont_embeddings.pooler_output
            # lm_output = self.my_head(cont_embeddings.last_hidden_state[:, 0, :])
            lm_output = cont_embeddings.last_hidden_state[:, 0, :]
            self.mylogger.debug(
                f"The output of the network looks like : {lm_output.shape}"
            )
            # cont_embeddings = self.lang_head(batch).hidden_states[-1]
            contextualized_embeddings_batches.append(lm_output)

        # Once We have embeddings we can compute the loss
        loss = self.loss(contextualized_embeddings_batches)
        loss_avg = loss.mean()  # Average across batches
        self.log(
            "train_loss", loss_avg.item(), prog_bar=True, on_step=True, on_epoch=True
        )
        return loss_avg

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-5)
        # Warmup scheduler
        # scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
        # optimizer, T_max=10, eta_min=0
        # )
        # return [optimizer], [{"scheduler": scheduler, "interval": "step"}]
        return optimizer

    def init_validation_tqdm(self):
        bar = super().init_validation_tqdm()
        bar.set_descripion("Validation")
        return bar

        # def validation_step(self, ref_batches: List[Tensor], batch_idx):
        # Get embedding
        # self.mylogger.debug(f"ref_batches: {ref_batches}")
        cont_embeddings = []
        # for batch in ref_batches:
        #    attention_mask = torch.ones_like(batch)
        #    attention_mask[batch == 0] = 0
        #    docs_embeds = self.lang_head(
        #        batch, attention_mask=attention_mask
        #    )  # .hidden_states[-1]
        #    # cont_embeddings.append(docs_embeds[:, 0, :])
        ## Once We have embeddings we can compute the loss
        # loss = self.loss(0.1)
        # loss_avg = loss.mean()
        # self.log("val/loss", loss_avg.item())
        # return loss_avg



import logging
from typing import List

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from transformers import PreTrainedTokenizer

from .utils.general import setup_logger


class MarginilizedLoss(nn.Module):
    def __init__(
        self, tokenizer: PreTrainedTokenizer, margin=0.1, hyp1=1, hkyp2=2, hyp3=3
    ):
        super(MarginilizedLoss, self).__init__()
        self.tokenizer = tokenizer  # 🪲
        self.logger = setup_logger(
            "MarginilizedLoss", "MarginilizedLoss.log", level=logging.INFO
        )

    def forward(self, x: List[Tensor]):
        """
        Will receive three different samples and will compute their loss
        based on the distance.
        args:
            x : List[Tensor]. The first n-1 tensors are positive samples whereas the last one is the negative sample.
                ref x batch_size x embedding_dim
        """
        # self.logger.debug(f"input x for loss is x: {x}")
        # intra_distance = torch.sum(
        #    torch.stack([self._distance(x[0], neighbor) for neighbor in x[1:-1]]), dim=0
        # )

        # self.logger.debug(f"input x for loss is x: {x}")
        # self.logger.debug(f"Shape of one of those is {x[0].shape}")
        intra_distance = self._distance(x[0], x[1])
        self.logger.debug(f"For samples : We get intra_distance: {intra_distance}")
        # self.logger.debug(f"intra_distance: {intra_distance}")
        inter_distance = self._distance(x[0], x[-1])
        self.logger.debug(f"For samples : We get inter_distance: {inter_distance}")
        # self.logger.debug(f"inter_distance: {inter_distance}")
        # difference = intra_distance - inter_distance + 1 * len(x[:-1])
        difference = (
            intra_distance - inter_distance + 1
        )  # I dont know I think this 10 might make sense
        self.logger.debug(f"difference: {difference}")
        zeros_like = torch.zeros_like(difference)
        # final_res = torch.max(torch.stack([difference, zeros_like]), dim=0).values
        final_res = torch.clamp(difference, min=0)
        self.logger.debug(
            f"Final result is :\n\t{final_res} with average {final_res.mean().item()}"
        )
        return final_res

    def _distance(self, x: Tensor, y: Tensor) -> Tensor:
        # epsilon = 1e-10
        # return torch.sqrt(
        # torch.sum(torch.pow(x - y, 2), dim=-1) + epsilon
        # )  # Returns unit [x] tensor
        return torch.sum(torch.pow(x - y, 2), dim=-1)  # Returns unit [x] tensor

    def _cosine_distance(self, x: Tensor, y: Tensor) -> Tensor:
        return torch.cosine_similarity(x, y, dim=-1)



"""
Written by @ottersome - Luis Garcia
Will create embeddings in vector space out of documents
"""
import logging
from typing import List

import torch
import torch.nn as nn
from torch import Tensor

from .utils.general import setup_logger


class DocumentEmbeddor(nn.Module):
    def __init__(self, embedding_dim, lh_model: nn.Module):
        super(DocumentEmbeddor, self).__init__()

        self.mylogger = setup_logger(
            "DocumentEmbeddor", "DocumentEmbeddor.log", level=logging.INFO
        )

        self.language_head = lh_model
        self.embedding_dim = embedding_dim

    def forward(self, token_batches: List[Tensor]):
        """
        For convenience we will have the forward pass take a tuple of papers
        and return their emebeddings. We will handle loss outside of the model
        args:
            tokens: List[Tensor] where each element might be a batched tensor
        """
        contextualized_embeddings_batches = []
        for batch in token_batches:
            # self.mylogger.debug(f"batch: {batch}")
            contextualized_batch = self.language_head(batch).hidden_states[
                -1
            ]  # Last hidden layer is at end of tuple

            # print(f"contextualized_batch: {contextualized_batch}")
            # TODO: we might want to add a pooling layer here or a linear layer to reduce the dimensionality

            # HACK : we have to be careful not to be averaging over all-zero vectors
            # Each element in the batch will be embedding_dim*num_embeddings. We want to average over num_embeddings:
            # contextualized_embeddings_batches.append(
            #    torch.mean(contextualized_batch, dim=1)
            # )
            # Get the first token ([CLS]) of each batch
            contextualized_embeddings_batches.append(contextualized_batch[:, 0, :])
        return contextualized_embeddings_batches



from setuptools import find_packages, setup

setup(
    name="llm4bi_embedder",
    version="0.1",
    packages=find_packages(),
    install_requires=[
        "lightning",
        "transformers",
        "pandas",
        "psycopg2-binary",
        "google-cloud-storage",
        "wandb",
        "pyarrow",
        "fastparquet",
        "sqlalchemy",
    ],
    author="Luis Garcia",
    author_email="admin@huginns.io",
    description="Package including any embedder for the LLM4BI project",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="GNU GPLv3",
    url="git@github.com:ottersome/patent_embedder.git",
)
